#version 430

#define M_PI 3.1415926535897932384626433832795

#define TYPE_EMITTER 0
#define TYPE_FLAME1 1
#define TYPE_FLAME2 2
#define TYPE_FLAME3 3
#define TYPE_SPARK 4

layout (local_size_x = 16, local_size_y = 16, local_size_y = 1) in;


struct ParticleData {
	vec4 position; // Vertex position & TTL
	vec4 velocity; // Vertex velocity & Start time
	vec4 rotScale; // x = Rotation, yz = Scale, w = unused
};

layout(std430, binding = 0) buffer ParticleData1 {
	ParticleData ParticleDataIncoming[];
};

layout(std430, binding = 1) buffer Type1 {
	uint TypesIncoming[];
};

layout(std430, binding = 2) buffer ParticleData2 {
	ParticleData ParticleDataOutgoing[];
};

layout(std430, binding = 3) buffer Type2 {
	uint TypesOutgoing[];
};

layout (binding = 4, offset = 0) uniform atomic_uint CurrentParticleCount;

uniform uint  LastParticleCount;
uniform uint  MaximumParticleCount;
uniform float DeltaT;
uniform float CurrentTime;
uniform uint  SpawnFlameParticleCount;

// Gravity
const vec3 GRAVITY = vec3(0, -9.81f, 0);

// From: http://stackoverflow.com/questions/4200224/random-noise-functions-for-glsl
// Returns pseudo-random numbers between (0, 1)
float rand(vec2 co)
{
	return fract(sin(dot(co.xy ,vec2(12.9898,78.233))) * 43758.5453);
}

float randRange(vec2 co, float minVal, float maxVal)
{
	return rand(co) * (maxVal - minVal) + minVal;
}

// From: http://www.neilmendoza.com/glsl-rotation-about-an-arbitrary-axis/
// axis is assumed to be normalized
mat3 rotationMatrix(vec3 axis, float angle)
{
	//axis = normalize(axis);
	float s = sin(angle);
	float c = cos(angle);
	float oc = 1.0 - c;

	return mat3(oc * axis.x * axis.x + c,           oc * axis.x * axis.y - axis.z * s,  oc * axis.z * axis.x + axis.y * s,
				oc * axis.x * axis.y + axis.z * s,  oc * axis.y * axis.y + c,           oc * axis.y * axis.z - axis.x * s,
				oc * axis.z * axis.x - axis.y * s,  oc * axis.y * axis.z + axis.x * s,  oc * axis.z * axis.z + c           );
	
	//return mat4(oc * axis.x * axis.x + c,           oc * axis.x * axis.y - axis.z * s,  oc * axis.z * axis.x + axis.y * s,  0.0,
	//			oc * axis.x * axis.y + axis.z * s,  oc * axis.y * axis.y + c,           oc * axis.y * axis.z - axis.x * s,  0.0,
	//			oc * axis.z * axis.x - axis.y * s,  oc * axis.y * axis.z + axis.x * s,  oc * axis.z * axis.z + c,           0.0,
	//			0.0,                                0.0,                                0.0,                                1.0);
}

void addParticle(vec3 position, vec3 velocity, float rotation, vec2 scaleXY, float ttl, float startTime, uint type)
{
	uint particleNr = atomicCounterIncrement(CurrentParticleCount);  // increments the counter and returns its prior value
	if (particleNr >= MaximumParticleCount)
	{
		// we are out of space
		atomicCounterDecrement(CurrentParticleCount);
		return;
	}

	ParticleDataOutgoing[particleNr] = ParticleData(
		vec4(position, ttl),
		vec4(velocity, startTime),
		vec4(rotation, scaleXY, 0)
	);
	TypesOutgoing[particleNr] = type;
}

void main()
{
	// unique one-dimensional index:
	uint index = gl_GlobalInvocationID.x + gl_GlobalInvocationID.y * gl_NumWorkGroups.x * gl_WorkGroupSize.x;
	if(index >= LastParticleCount) return;

	// Read data:
	ParticleData pd = ParticleDataIncoming[index];
	vec4 d1 = pd.position;
	vec4 d2 = pd.velocity;
	vec4 d3 = pd.rotScale;

	vec3 position = d1.xyz;
	vec3 velocity = d2.xyz;
	float ttl = d1.w;
	float startTime = d2.w;
	uint type = TypesIncoming[index];
	float rotation = d3.x;
	vec2 scaleXY = d3.yz;

	if (startTime > CurrentTime)
	{
		// wait until its time for the particle
		addParticle(position, velocity, rotation, scaleXY, ttl, startTime, type);
		return;
	}

	// Update time-to-live:
	ttl -= DeltaT;
	if (ttl <= 0) return; // end of live for this particle

	if (type == TYPE_EMITTER)
	{
		addParticle(position, velocity, rotation, scaleXY, ttl, startTime, type);

		if (SpawnFlameParticleCount > 0)
		{
			// spawn new flames, yeah!
			for (int i = 0; i < SpawnFlameParticleCount; i++)
			{
				// spawn in box (box emitter):
				float r1 = rand(gl_GlobalInvocationID.xy + vec2(i * 2.3, i * 3.6) + vec2(mod(CurrentTime, 1.2)));
				float r2 = rand(gl_GlobalInvocationID.xy + vec2(i * 3.1, i * 2.4) + vec2(mod(CurrentTime, 6.8)));
				float r3 = rand(gl_GlobalInvocationID.xy + vec2(i * 4.2, i * 3.7) + vec2(mod(CurrentTime, 10.7)));
				float r4 = rand(gl_GlobalInvocationID.xy + vec2(i * 1.4, i * 7.1) + vec2(mod(CurrentTime, 4.2)));
				float r5 = rand(gl_GlobalInvocationID.xy + vec2(i * 2.1, i * 3.4) + vec2(mod(CurrentTime, 9.8)));

				float x = r1 * 2 - 1; // between (-1, 1)
				float y = 0;
				float z = r3 * 2 - 1;

				vec3 positionNew = position + vec3(x, y, z) * 0.01;

				// TODO random rotation and scaling and velocity?
				// TODO initially small but grows fast to a maximum size
				// TODO use different types of flames
				// TODO at some point it becomes smoke

				float initialRotation = r2 * 2 * M_PI;
				float flameTTL = r4 * 0.5 + 1.2;
				vec3 initalVelocity = vec3(0, 0.01, 0) * r3;
				vec2 initialSize = vec2(0.01, 0.01) * r1;
				uint type = uint((TYPE_FLAME3 - TYPE_FLAME1) * r5 + TYPE_FLAME1 + 0.5);
				addParticle(positionNew, initalVelocity, initialRotation, initialSize, flameTTL, CurrentTime, type);
			}
		}
	}
	else
	{
		// Calculate movement:

		// 1. Calculate forces:
		vec3 force = vec3(0, 0.1, 0); // slowly rising

		// add wind:
		float windX = pow(sin(CurrentTime) * 0.5 + 0.5, 4) * 150;
		float windZ = sin(CurrentTime * 100) * 20;
		force += DeltaT * vec3(windX, 0, windZ) / 75;

		// 2. Calculate acceleration:
		vec3 acc = force;

		// 3. Calculate new velocity:
		velocity += acc * DeltaT;

		// 4. Calculate new position:
		position += velocity * DeltaT;

		rotation += 0.2 * DeltaT;

		// particle grows
		const vec2 maxSize = vec2(0.05, 0.05);
		scaleXY = min(scaleXY + 0.05 * vec2(DeltaT), maxSize);

		addParticle(position, velocity, rotation, scaleXY, ttl, startTime, type);
	}
}

