#version 330

// Adapted from: http://www.geeks3d.com/20140815/particle-billboarding-with-the-geometry-shader-glsl/

layout (points) in;
layout (triangle_strip, max_vertices = 4) out;

uniform mat4 ModelMatrix;   
uniform mat4 ModelViewMatrix;
uniform mat4 ProjectionMatrix;

uniform float CurrentTime;

in vData
{
	vec4 Velocity1;
	vec4 RotScale1;
	uint Type1;
	float TTL1;
} vertex[];

out vec2 TexCoord0;
flat out vec4 Velocity0;
flat out uint Type0;
flat out float TTL0;

/*
// quad will be rotated according to velocity vector
void emitRotatedQuad()
{
	vec4 P = gl_in[0].gl_Position;
	vec4 velocity = vertex[0].Velocity1;
	vec3 direction = normalize(velocity.xyz);
	vec3 right = vec3(1, 0, 0);
	vec3 some = cross(right, direction);
	right = cross(direction, some);

	right = right * 0.5 * ParticleSize;
	direction = direction * 0.5 * ParticleSize;


	// a: left-bottom
	vec3 va = P.xyz - right - direction;
	gl_Position = ProjectionMatrix * vec4(va, P.w);
	TexCoord0 = vec2(0.0, 0.0);
	Velocity0 = vertex[0].Velocity1;
	TTL0 = vertex[0].TTL1;
	Type0 = vertex[0].Type1;
	EmitVertex();

	// b: right-bottom
	vec3 vb = P.xyz + right - direction;
	gl_Position = ProjectionMatrix * vec4(vb, P.w);
	TexCoord0 = vec2(1.0, 0.0);
	Velocity0 = vertex[0].Velocity1;
	TTL0 = vertex[0].TTL1;
	Type0 = vertex[0].Type1;
	EmitVertex();


	// c: left-top
	vec3 vc = P.xyz - right + direction;
	gl_Position = ProjectionMatrix * vec4(vc, P.w);
	TexCoord0 = vec2(0.0, 1.0);
	Velocity0 = vertex[0].Velocity1;
	TTL0 = vertex[0].TTL1;
	Type0 = vertex[0].Type1;
	EmitVertex();

	// d: right-top
	vec3 vd = P.xyz + right + direction;
	gl_Position = ProjectionMatrix * vec4(vd, P.w);
	TexCoord0 = vec2(1.0, 1.0);
	Velocity0 = vertex[0].Velocity1;
	TTL0 = vertex[0].TTL1;
	Type0 = vertex[0].Type1;
	EmitVertex();
}
*/

// standard facing quad -> orientation is dependent on camera rotation
// = view plane alignment
void emitViewPlaneFacingQuad(vec2 size, float rotation)
{
	vec4 P = gl_in[0].gl_Position;

	float cosr = cos(rotation);
	float sinr = sin(rotation);
	mat2 rotMat = mat2(cosr, sinr, -sinr, cosr);

	// a: left-bottom 
	vec2 va = P.xy + rotMat * vec2(-0.5, -0.5) * size;
	gl_Position = ProjectionMatrix * vec4(va, P.zw);
	TexCoord0 = vec2(0.0, 0.0);
	Velocity0 = vertex[0].Velocity1;
	TTL0 = vertex[0].TTL1;
	Type0 = vertex[0].Type1;
	EmitVertex();
  
	// b: right-bottom
	vec2 vd = P.xy + rotMat * vec2(0.5, -0.5) * size;
	gl_Position = ProjectionMatrix * vec4(vd, P.zw);
	TexCoord0 = vec2(1.0, 0.0);
	Velocity0 = vertex[0].Velocity1;
	TTL0 = vertex[0].TTL1;
	Type0 = vertex[0].Type1;
	EmitVertex();

	// c: left-top
	vec2 vb = P.xy + rotMat * vec2(-0.5, 0.5) * size;
	gl_Position = ProjectionMatrix * vec4(vb, P.zw);
	TexCoord0 = vec2(0.0, 1.0);
	Velocity0 = vertex[0].Velocity1;
	TTL0 = vertex[0].TTL1;
	Type0 = vertex[0].Type1;
	EmitVertex();
  
	// d: right-top
	vec2 vc = P.xy + rotMat * vec2(0.5, 0.5) * size;
	gl_Position = ProjectionMatrix * vec4(vc, P.zw);
	TexCoord0 = vec2(1.0, 1.0);
	Velocity0 = vertex[0].Velocity1;
	TTL0 = vertex[0].TTL1;
	Type0 = vertex[0].Type1;
	EmitVertex();

	EndPrimitive();
}


// From: http://www.neilmendoza.com/glsl-rotation-about-an-arbitrary-axis/
// axis is assumed to be normalized
mat3 rotationMatrix(vec3 axis, float angle)
{
	//axis = normalize(axis);
	float s = sin(angle);
	float c = cos(angle);
	float oc = 1.0 - c;

	return mat3(oc * axis.x * axis.x + c,           oc * axis.x * axis.y - axis.z * s,  oc * axis.z * axis.x + axis.y * s,
				oc * axis.x * axis.y + axis.z * s,  oc * axis.y * axis.y + c,           oc * axis.y * axis.z - axis.x * s,
				oc * axis.z * axis.x - axis.y * s,  oc * axis.y * axis.z + axis.x * s,  oc * axis.z * axis.z + c           );
	
	//return mat4(oc * axis.x * axis.x + c,           oc * axis.x * axis.y - axis.z * s,  oc * axis.z * axis.x + axis.y * s,  0.0,
	//			oc * axis.x * axis.y + axis.z * s,  oc * axis.y * axis.y + c,           oc * axis.y * axis.z - axis.x * s,  0.0,
	//			oc * axis.z * axis.x - axis.y * s,  oc * axis.y * axis.z + axis.x * s,  oc * axis.z * axis.z + c,           0.0,
	//			0.0,                                0.0,                                0.0,                                1.0);
}

// quad faces the camera location and is along the given up vector
// = axial billboard
void emitViewPointFacingQuad(vec2 size, vec3 up, float rotationAngle)
{
	vec4 P = gl_in[0].gl_Position;
	vec3 P3 = P.xyz;

	up = normalize(ModelViewMatrix * vec4(up, 0)).xyz; // up vector in world space -> to view space
	vec3 direction = normalize(P3); // from camera to point
	vec3 right = cross(direction, up);

	mat3 rotMat = rotationMatrix(direction, rotationAngle);
	right = rotMat * right;
	up = rotMat * up;

	right = right * 0.5 * size.x;
	up = up * 0.5 * size.y;

	// a: left-bottom
	vec3 va = P3 - right - up;
	gl_Position = ProjectionMatrix * vec4(va, P.w);
	TexCoord0 = vec2(0.0, 0.0);
	Velocity0 = vertex[0].Velocity1;
	TTL0 = vertex[0].TTL1;
	Type0 = vertex[0].Type1;
	EmitVertex();

	// b: right-bottom
	vec3 vb = P3 + right - up;
	gl_Position = ProjectionMatrix * vec4(vb, P.w);
	TexCoord0 = vec2(1.0, 0.0);
	Velocity0 = vertex[0].Velocity1;
	TTL0 = vertex[0].TTL1;
	Type0 = vertex[0].Type1;
	EmitVertex();


	// c: left-top
	vec3 vc = P3 - right + up;
	gl_Position = ProjectionMatrix * vec4(vc, P.w);
	TexCoord0 = vec2(0.0, 1.0);
	Velocity0 = vertex[0].Velocity1;
	TTL0 = vertex[0].TTL1;
	Type0 = vertex[0].Type1;
	EmitVertex();

	// d: right-top
	vec3 vd = P3 + right + up;
	gl_Position = ProjectionMatrix * vec4(vd, P.w);
	TexCoord0 = vec2(1.0, 1.0);
	Velocity0 = vertex[0].Velocity1;
	TTL0 = vertex[0].TTL1;
	Type0 = vertex[0].Type1;
	EmitVertex();
}

void main (void)
{
	float particleStartTime = vertex[0].Velocity1.w;
	if (CurrentTime < particleStartTime) return; // don't draw particles with a future start time

	uint type = vertex[0].Type1;
	vec2 scaling = vertex[0].RotScale1.yz;
	float rotation = vertex[0].RotScale1.x;

	if (type == 0u)
	{
		// emitter -> should not be drawn??
		vec2 size2 = (ModelMatrix * vec4(scaling, 0, 0)).xy;
		emitViewPlaneFacingQuad(size2, 0); // round particles should always face the camera
	}
	else if (type >= 1u)
	{
		// flame
		vec3 up = vertex[0].Velocity1.xyz;
		//vec3 up = vec3(0, 1, 0);
		vec2 size2 = (ModelMatrix * vec4(scaling, 0, 0)).xy;
		//emitViewPointFacingQuad(size2, up, rotation);
		emitViewPlaneFacingQuad(size2, rotation);
	}
}
