#version 430

#define M_PI 3.1415926535897932384626433832795

#define TYPE_START 0
#define TYPE_EXPLOSION1 1

#define UNTRAILABLE_START 2
	#define TRAILS_START UNTRAILABLE_START
		#define TYPE_TRAIL1 TRAILS_START + 0
		#define TYPE_TRAIL2 TRAILS_START + 1
		#define TYPE_TRAIL3 TRAILS_START + 2
		#define TYPE_TRAIL4 TRAILS_START + 3
	#define TRAILS_END TYPE_TRAIL4
	#define TYPE_SMOKE UNTRAILABLE_START + 4
#define UNTRAILABLE_END UNTRAILABLE_START + 5


layout (local_size_x = 16, local_size_y = 16, local_size_y = 1) in;


// Target 0 : Vertex position & TTL
layout(std430, binding = 0) buffer Pos1 {
   vec4 PositionsIncoming[];
};

// Target 1 : Vertex velocity & Start time
layout(std430, binding = 1) buffer Vel1 {
	vec4 VelocitiesIncoming[];
};

layout(std430, binding = 2) buffer Type1 {
	uvec4 TypesIncoming[];
};

layout(std430, binding = 3) buffer Pos2 {
   vec4 PositionsOutgoing[];
};

layout(std430, binding = 4) buffer Vel2 {
	vec4 VelocitiesOutgoing[];
};

layout(std430, binding = 5) buffer Type2 {
	uvec4 TypesOutgoing[];
};

layout (binding = 6, offset = 0) uniform atomic_uint CurrentParticleCount;

uniform uint  LastParticleCount;
uniform uint  MaximumParticleCount;
uniform float DeltaT;
uniform float CurrentTime;
uniform uint  SpawnTrailParticleCount;

// Gravity
const vec3 GRAVITY = vec3(0, -9.81f, 0);

// From: http://stackoverflow.com/questions/4200224/random-noise-functions-for-glsl
// Returns pseudo-random numbers between (0, 1)
float rand(vec2 co)
{
	return fract(sin(dot(co.xy ,vec2(12.9898,78.233))) * 43758.5453);
}

float randRange(vec2 co, float minVal, float maxVal)
{
	return rand(co) * (maxVal - minVal) + minVal;
}

// From: http://www.neilmendoza.com/glsl-rotation-about-an-arbitrary-axis/
// axis is assumed to be normalized
mat3 rotationMatrix(vec3 axis, float angle)
{
	//axis = normalize(axis);
	float s = sin(angle);
	float c = cos(angle);
	float oc = 1.0 - c;

	return mat3(oc * axis.x * axis.x + c,           oc * axis.x * axis.y - axis.z * s,  oc * axis.z * axis.x + axis.y * s,
				oc * axis.x * axis.y + axis.z * s,  oc * axis.y * axis.y + c,           oc * axis.y * axis.z - axis.x * s,
				oc * axis.z * axis.x - axis.y * s,  oc * axis.y * axis.z + axis.x * s,  oc * axis.z * axis.z + c           );
	
	//return mat4(oc * axis.x * axis.x + c,           oc * axis.x * axis.y - axis.z * s,  oc * axis.z * axis.x + axis.y * s,  0.0,
	//			oc * axis.x * axis.y + axis.z * s,  oc * axis.y * axis.y + c,           oc * axis.y * axis.z - axis.x * s,  0.0,
	//			oc * axis.z * axis.x - axis.y * s,  oc * axis.y * axis.z + axis.x * s,  oc * axis.z * axis.z + c,           0.0,
	//			0.0,                                0.0,                                0.0,                                1.0);
}

void addParticle(vec3 position, vec3 velocity, float ttl, float startTime, uint type, uvec3 color)
{
	uint particleNr = atomicCounterIncrement(CurrentParticleCount);  // increments the counter and returns its prior value
	if (particleNr >= MaximumParticleCount)
	{
		// we are out of space
		atomicCounterDecrement(CurrentParticleCount);
		return;
	}
	PositionsOutgoing[particleNr] = vec4(position, ttl);
	VelocitiesOutgoing[particleNr] = vec4(velocity, startTime);
	TypesOutgoing[particleNr] = uvec4(color, type);
}

void main()
{
	// unique one-dimensional index:
	uint index = gl_GlobalInvocationID.x + gl_GlobalInvocationID.y * gl_NumWorkGroups.x * gl_WorkGroupSize.x;
	if(index >=	 LastParticleCount) return;

	// Read data:
	vec4 d1 = PositionsIncoming[index];
	vec4 d2 = VelocitiesIncoming[index];
	uvec4 d3 = TypesIncoming[index];

	vec3 position = d1.xyz;
	vec3 velocity = d2.xyz;
	uvec3 color = d3.xyz;
	float ttl = d1.w;
	float startTime = d2.w;
	uint type = d3.w;
	bool isTrail = (type >= UNTRAILABLE_START) && (type < UNTRAILABLE_END);

	if (startTime > CurrentTime)
	{
		// wait until its time for the particle
		addParticle(position, velocity, ttl, startTime, type, color);
		return;
	}


	// Calculate movement:
	// 1. Calculate forces:
	vec3 force = GRAVITY;
	//force += (0.5f, 0, 0); // add wind

	if (isTrail)
	{
		// lets pretend the trails are very light and only fall very slowly
		force *= 0.1;
	}

	// 2. Calculate acceleration:
	vec3 acc = force;

	// 3. Calculate new velocity:
	velocity += acc * DeltaT;

	// 4. Calculate new position:
	position += velocity * DeltaT;
	

	vec3 velNorm = normalize(velocity);
	vec3 velOrtho = cross(velNorm, vec3(1, 0, 0));


	if (type == TYPE_START)
	{
		// add rotational element to movement:
		const float amplitude = 10;
		const float frequency = 15;

		mat3 rotMat = rotationMatrix(velNorm, (CurrentTime - startTime) * frequency);
		vec3 some = rotMat * velOrtho;

		position += some * DeltaT * amplitude;
	}

	// Update time-to-live:
	ttl -= DeltaT;

	if (ttl > 0)
	{
		// particle still alive -> add to output list
		addParticle(position, velocity, ttl, startTime, type, color);

		// spawn trail particles:
		if (!isTrail && SpawnTrailParticleCount > 0)
		{
			// create new trail particles:
			float mixT = clamp(ttl / 1.5, 0.5, 1.0);
			int toSpawn = int(SpawnTrailParticleCount * mixT);
			for (int i = 0; i < toSpawn; i++)
			{
				const float offset = randRange(gl_GlobalInvocationID.xy + vec2(3.9 + i, 6.4), 1.0, 2.0); // offset from particle position
				const float radius = randRange(gl_GlobalInvocationID.xy + vec2(2.3 + i, 3.4), 0.5, 1); // radius of rotational position
				const float velocityFraction = 0.1;

				float ttl1 = randRange(gl_GlobalInvocationID.xy + vec2(2.3 + i, 3.4), 0.8, 2.5);

				float rotAngle = randRange(position.xy + vec2(i * 1.2, i * 3.4), -M_PI, M_PI);
				mat3 rotMat = rotationMatrix(velNorm, rotAngle);
				

				vec3 some = rotMat * velOrtho;
				some *= radius;

				vec3 newPos = position - offset * velNorm + some;
				if (type == TYPE_START)
				{
					addParticle(newPos, velocity * velocityFraction, ttl1, CurrentTime, TYPE_TRAIL1, uvec3(255, 255, 255));
				}
				else
				{
					addParticle(newPos, velocity * velocityFraction, ttl1, CurrentTime, TYPE_TRAIL1, color);
				}
			}
		}
	}
	else
	{
		if (type == TYPE_START)
		{
			// explode!!
			const int newParticleCount = 100;

			// calculate particle color by rotating in HSV space and converting to RGB
			// see http://www.rapidtables.com/convert/color/hsv-to-rgb.htm
			const float H = mod(CurrentTime * 20, 275) + 90; // color (only between 90 and 360)
			const float S = 0.75f; // saturation
			const float V = 0.75f; // brightness

			const float C = V * S;
			const float X = C * (1 - abs(mod(H / 60, 2) - 1));
			const float m = V - C;

			float R = 0; float G = 0; float B = 0;
			if (H >= 0 && H < 60)
			{
				R = C; B = X; G = 0;
			}
			else if (H < 120)
			{
				R = X; B = C; G = 0;
			}
			else if (H < 180)
			{
				R = 0; B = C; G = X;
			}
			else if (H < 240)
			{
				R = 0; B = X; G = C;
			}
			else if (H < 300)
			{
				R = X; B = 0; G = C;
			}
			else
			{
				R = C; B = 0; G = X;
			}

			const uvec3 flareParticleColor = uvec3((vec3(R, G, B) + m) * 255.0f);

			for (int i = 0; i < newParticleCount; i++)
			{
				// create random vectors on the unit sphere
				float r1 = rand(gl_GlobalInvocationID.xy + vec2(i * 2.3, i * 3.6));
				float r2 = rand(gl_GlobalInvocationID.xy + vec2(i * 3.1, i * 2.4));

				float theta = 2 * r2 * M_PI - M_PI; // between (-pi, pi)

				float z = 1.5 * r1 - 0.5; // between (-0.5, 1)			full sphere: 2 * r1 - 1; // between (-1, 1)
				float x = sin(theta) * sqrt(1 - z * z);
				float y = cos(theta) * sqrt(1 - z * z);
				// x^2 + y^2 = 1 - z^2   ->   x^2 + y^2 + z^2 = 1

				const float explosionSpeed = 30.0f;
				vec3 velocityNew = vec3(x, z, y) * explosionSpeed;
				const float ttl = 3.5;
				addParticle(position, velocityNew, ttl, CurrentTime, TYPE_EXPLOSION1, flareParticleColor);
			}
		}
	}
}

